module spiderweb

	using IntervalArithmetic, Polynomials, LinearAlgebra

	export newton, radii, proof_𝑛ℓ



	function proof_𝑛ℓ(λ::Float64, 𝑛_range::Vector, ℓ_range::Vector)
		# Launch the proof for all 𝑛 ∈ 𝑛_range and ℓ ∈ ℓ_range
		m0 = 0. # set central mass to zero
		x = float.(1:𝑛_range[1])
		x2 = x
		for ℓ ∈ ℓ_range
		    for 𝑛 ∈ 𝑛_range
		        m = ones(Float64, 𝑛)./ℓ # set m_i = 1/ℓ for i > 0
		        if 𝑛 == 𝑛_range[1] && ℓ == ℓ_range[1] # initial guess for Newton's method
		            x = newton(x, m0, m, λ, 𝑛, ℓ)
		            radii(x, m0, m, λ, 𝑛, ℓ)
		            x2 = x # saving solution with the smallest number of circles
				elseif 𝑛 > 𝑛_range[1]
					push!(x, x[𝑛-1]^2/x[𝑛-2])
					x .*= range(.98, stop = .95, length = 𝑛) # correction for better convergence
					x = newton(x, m0, m, λ, 𝑛, ℓ)
					radii(x, m0, m, λ, 𝑛, ℓ)
		        elseif 𝑛 == 𝑛_range[1] && ℓ > ℓ_range[1]
		            x = x2[1:𝑛]
		            x .*= 0.8
		            x = newton(x, m0, m, λ, 𝑛, ℓ)
		            radii(x, m0, m, λ, 𝑛, ℓ)
		            x2 = x # saving solution with the smallest number of circles
		        end
		    end
		end
	end


	function newton(x::Vector{T}, m0::T, m::Vector{T}, λ::T, 𝑛::Int, ℓ::Int;
				max_iter = 50, atol = 1e-6, rtol = 1e-6) where {T<:Float64}
        # Newton iteration

        F, DF = F_DF(x, m0, m, λ, 𝑛, ℓ)

        isapprox(0., norm(F, Inf), atol=atol, rtol=rtol) && return x

        for i in 1:max_iter
            x -= DF\F
			x = sort(abs.(x))
            F, DF = F_DF(x, m0, m, λ, 𝑛, ℓ)
            isapprox(0., norm(F, Inf), atol=atol, rtol=rtol) && return x
        end
        error("No convergence after ",max_iter," iterations: |F| = ",norm(F, Inf),". Setup: atol = ",atol,", rtol = ",rtol,".")
	end


	function radii(x, m0, m, λ, 𝑛, ℓ)
		# Radii Polynomial Approach
		im0 = Interval(m0)
		iM = Interval.(m)
		ix = Interval.(x)
		r_star = 1e-5
		ixr = zeros(Interval{Float64}, 𝑛)
		@inbounds for i ∈ 1:𝑛
			ixr[i] = Interval(x[i] - r_star, x[i] + r_star)
		end

		iF, iDF, iDDF = iF_iDF_iDDF(ix, ixr, im0, iM, λ, 𝑛, ℓ)
		iA = inv(mid.(iDF))

		Y0 = sup(norm(iA*iF, Inf))
		Z0 = sup(opnorm(Matrix{Interval{Float64}}(I, 𝑛 ,𝑛) - iA*iDF, Inf))
		Z2 = sup(opnorm(iA*iDDF, Inf))

		r = roots(Poly([Y0, -(1-Z0), Z2]))
		if isreal(r)
			sort!(r)
			if r[1] ≤ r_star
				println("\033[1m\033[32m", "Success for (𝑛,ℓ) = (",𝑛,",",ℓ,") -> EI = [",r[1]," , ",min(r[2],r_star),"] (r* = ",r_star,").")
			else
				println("\033[1m\033[31m", "Failure for (𝑛,ℓ) = (",𝑛,",",ℓ,") -> smallest root is larger than r* = ",r_star,".
					Possible fix: 1) improve numerical approximation of the zero;
								  2) increase r*.")
			end
		else
			println("\033[1m\033[31m", "Failure for (𝑛,ℓ) = (",𝑛,",",ℓ,") -> complex roots.
				Possible fix: improve numerical approximation of the zero.")
		end
	end



	function F_DF(x, m0, m, λ, 𝑛, ℓ)
		# compute the map F whose zeros give a spiderweb central configuration
		# and its jacobian DF
		sqrt2 = sqrt(2)
		F = λ*x
		DF = Matrix{Float64}(I, 𝑛, 𝑛)*λ
		@inbounds for j ∈ eachindex(x)
			xj2 = x[j]^2
			for i ∈ eachindex(x)
				xi2, xixj = x[i]^2, x[i]*x[j]
				xi3 = xi2*x[i]

				if (j==1) F[i] += m0/xi2 end
				if (i==j) DF[i,j] += -2*m0/xi3 end

				for k ∈ 0:ℓ-1
					θ = k*360/ℓ
					cosdθ = cosd(θ)
					cosd2θ = cosd(2*θ)
					if i ≠ j
						d = (xi2 + xj2 - 2*xixj*cosdθ)^(1/2)
						d5 = 2*d^5
						F[i] += m[j]*(x[i] - x[j]*cosdθ)/d^3
						DF[i,i] += -m[j]*(4*xi2 + xj2 - 8*xixj*cosdθ + 3*xj2*cosd2θ)/d5
						DF[i,j] += -m[j]*(-4*(xi2 + xj2)*cosdθ + xixj*(7 + cosd2θ))/d5
					elseif i==j && k>0
						ζ = sqrt2*sqrt(1-cosdθ)
						F[i] += m[i]/(xi2*2*ζ)
						DF[i,j] += -m[i]/(xi3*ζ)
					end
				end
			end
		end
		return F, DF
	end



	function iF_iDF_iDDF(ix::Vector{T}, ixr, im0, iM, λ, 𝑛, ℓ) where {T<:Interval{Float64}}
		# compute in interval arithmetic the map F, its jacobian DF and the
		# hessian DDF (for the Z₂ bound)
		iπ = @interval(π)
		isqrt2 = @interval(sqrt(2))
		iF = λ*ix
		iDF = Matrix{T}(I, 𝑛, 𝑛)*λ
		iDDF  = zeros(T, 𝑛, 𝑛*𝑛)
		iDDFixjxj = zeros(T, 𝑛, 𝑛)
		@inbounds for j ∈ eachindex(ix)
			ixj2 = pow(ix[j],2)
			ixrj2 = pow(ixr[j],2)
			for i ∈ eachindex(ix)
				ixi2, ixixj = pow(ix[i],2), ix[i]*ix[j]
				ixi3 = ixi2*ix[i]
				ixri2, ixrixrj = pow(ixr[i],2), ixr[i]*ixr[j]
				ixri4 = ixri2*ixri2

				if (j==1) iF[i] += im0/ixi2 end
				if i==j
					iDF[i,j] += -2*im0/ixi3
					iDDFixjxj[i,j] = 6*im0/ixri4
				end

				for k ∈ 0:ℓ-1
					iθ = k*2*iπ/ℓ
					icosθ = cos(iθ)
					icos2θ = cos(2*iθ)
					if i!=j
						id = sqrt(ixi2 + ixj2 - 2*ixixj*icosθ)
						id5 = 2*pow(id,5)
						iF[i] += iM[j]*(ix[i] - ix[j]*icosθ)/pow(id,3)
						iDF[i,i] += -iM[j]*(4*ixi2 + ixj2 - 8*ixixj*icosθ + 3*ixj2*icos2θ)/id5
						iDF[i,j] += -iM[j]*(-4*(ixi2 + ixj2)*icosθ + ixixj*(7 + icos2θ))/id5
						idr = sqrt(ixri2 + ixrj2 - 2*ixrixrj*icosθ)
						idr7 = 2*pow(idr,7)
						iDDFixjxj[i,i] += 3*iM[j]*(ixr[i] - ixr[j]*icosθ)*(4*ixri2 - ixrj2 - 8*ixrixrj*icosθ + 5*ixrj2*icos2θ)/idr7
						iDDFixjxj[i,j] += -3*iM[j]*(ixr[j]*(8*ixrj2 + 23*ixri2)*icosθ - ixr[i]*(20*ixrj2 + 2*ixri2 + (4*ixrj2 + 6*ixri2)*icos2θ - ixrixrj*cos(3*iθ)))/(2*idr7)
					elseif i==j && k>0
						iζ = isqrt2*sqrt(1-icosθ)
						iF[i] += iM[i]/(ixi2*2*iζ)
						iDF[i,j] += -iM[i]/(ixi3*iζ)
						iDDFixjxj[i,j] += 3*iM[i]/(ixri4*iζ)
					end
				end
			end
		end
		@inbounds for k ∈ eachindex(ix)
			s = (k-1)*𝑛
			iDDF[𝑛*s+1:𝑛+1:𝑛*s+𝑛*𝑛] = iDDFixjxj[k,1:𝑛]
			iDDF[k,1+s:𝑛+s] = iDDFixjxj[1:𝑛,k]
			iDDF[1:𝑛,k+s] = iDDFixjxj[1:𝑛,k]
		end
		return iF, iDF, iDDF
	end

end
